from xmldt import XmlDt


class HtmlDt (XmlDt):
    def __init__(self, **kwargs):
        super().__init__(**{**kwargs, "html": True})  # couldn't rewrite better

    def __call__(self, *args, **kwargs):
        super().__call__(*args, **kwargs)

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **{**kwargs, "html": True})






import sys

from xmldt import XmlDt
import argparse
import re


def build_skel(filename, name="DTSkel"):
    data = Skel(filename=filename)
    code = [f"""from xmldt import XmlDt
import sys


class {name} (XmlDt):

    def __default__(self, element):
          return f"[tag: {{element.tag}}, contents: {{element.contents}}]"
"""]
    for tag in data:
        method = re.sub("[.]", "_", tag)
        attr_count = str.join(" ", [f"{a}: {data[tag][a]}" for a in data[tag].keys() if a != ":count"])
        if method != tag:
            code.append(f"""    # @XmlDt.tag("{tag}")""")
        code.append(f"""    # def {method}(self, e): return e.tag # {data[tag][":count"]} occs   {attr_count}\n\n""")
    code.append(f"""
    
file = sys.argv[1]
dt = {name}()
print(dt(filename=file))\n""")
    return "\n".join(code)


def build_skel_jj(files, name="proc"):
    #skel = Skeljj(filename=filename) --->FIXME: dá None

    ### claro que sim. o __default__ nao tme return
    ### ver a minha solucao abaixo

    skel = Skeljj()
    for filename in files:
        skel(filename=filename)
    print(
f"""#!/usr/bin/python3
from xmldt import XmlDt, toxml
import sys
class {name} (XmlDt):
    pass

    # def __default__(s, e): 
    #     return f"[tag: {{e.tag}}, contents: {{e.c}}, attr: {{e.v}}]"
""")
    done = set()
    for tag in skel.order:
        v=skel.data[tag]
        attrs = str.join(" ", [f"{a}({v[a]})" for a in v.keys()-{":count",":pcdata"}])
        if ":pcdata" in v:
            pcdatalen = v[':pcdata'] // v[':count']
            cont= f"{v[':count']} {attrs} PCDATAlen({pcdatalen})"
        else:
            cont= f"{v[':count']} {attrs}" 
        indent="   "*(len(skel.ans[tag]))

        print( f"    # def {indent}{tag}(s, e):   # {cont}" )
#        for sun in skel.suns.get(tag,[]):
#            print( f"    # ...    {indent}{sun}")
    print()
    print("file = sys.argv[1]")
    print(f"print({name}(filename=file))")

#    for t in (skel.pathorder):
#        print(":".join(t))

def ensurepath(seq,path):
    for i in range(len(path)):
        t = tuple(path[:i+1])
        if t not in seq:
            seq.append(t)

class Skeljj (XmlDt):

    data={}              ## tag → (attr | # → int)
    ans={}     ## ansesters tag → tag-path  (of first oco)
    suns={}              ## tag → set(tag)
    order=[]             ## list(tag)
    pathorder=[]

    def __pcdata__(self, e):
        if e.isspace(): return ""
        self.path[-1][":pcdata"] = (self.path[-1][":pcdata"] or 0) + len(e) 
        print(f"DEBUG {e} {len(e)}")
        
#
#        if f := self.path[-1] :
#            self.data.setdefault(f.tag,{":count":0})
#            if "@pcdata" not in self.data[f.tag]:
#                self.data[f.tag]["@pcdata"] = 0
#            self.data[f.tag]["@pcdata"] += len(e)

    def __default__(self, e):
        path = [a.tag for a in self.path] + [e.tag]
#        ensurepath(self.pathorder, path)
#        if f := e.father :
#            self.suns.setdefault(f.tag,{})[e.tag] = 1
        if e.tag not in self.data:
#            self.data[e.tag] = {":count": 0}
            p = self.ans[e.tag]=[a.tag for a in self.path]
            for tag in p+[e.tag]:
                if tag in self.order: continue 
                self.order.append(tag)
            self.data[e.tag] = {":count": 0}
        self.data[e.tag][":count"] += 1
        for key in e.attrs.keys():
            if key not in self.data[e.tag]:
                self.data[e.tag][key] = 0
            if key == ":pcdata":
                self.data[e.tag][key] += e[":pcdata"]  ## len of text
            else:
                self.data[e.tag][key] += 1


class Skel (XmlDt):

    data = {}

    def __default__(self, element):
        if element.tag not in self.data:
            self.data[element.tag] = {":count": 0}
        self.data[element.tag][":count"] += 1
        for key in element.attrs.keys():
            if key not in self.data[element.tag]:
                self.data[element.tag][key] = 0
            self.data[element.tag][key] += 1

    def __end__(self, _):
        return self.data


def main():
    arg_parser = argparse.ArgumentParser()
    skel = arg_parser.add_argument_group()
    skel.add_argument("-s", "--skel", help="Creates basic skeleton file", action="store_true")
    skel.add_argument("filename", type=str, nargs='+', help="the XML file to use as model")
    args = arg_parser.parse_args()

    if args.skel:
        print(build_skel(args.filename[0]))
    else:
        build_skel_jj(args.filename)



"""
  python module to Down Translate XML

Synopsis
========

# Description

"""
import pprint

from lxml.etree import XMLParser, parse, fromstring, Comment, HTMLParser
from xmldt.element import Element, toxml
import sys

__version__ = "0.0.4"
__docformat__ = 'markdown'


class XmlDt:

    @classmethod
    def tag(cls, name):
        def decorator(func):
            def tmp(*args, **kwargs):
                return func(*args, **kwargs)

            tmp.has_alias = name
            return tmp
        return decorator

    def __init__(self, strip=False, empty=False, html=False):
        self._parser = XMLParser(recover=True) if not html else HTMLParser(recover=True)
        self._path = []
        self._strip = strip
        self._empty = empty
        self._types = getattr(self, "_types", {})
        self._alias = {
            method.has_alias: method for method in {
                getattr(self, name) for name in dir(self)
                if callable(getattr(self, name)) and hasattr(getattr(self, name), "has_alias")}}

    def __new__(cls, xml=None, filename=None, strip=False, empty=False, html=False):
        # nao gosto, mas funciona
        self = super(XmlDt, cls).__new__(cls)
        self.__init__(strip=strip, empty=empty, html=html)

        if filename is not None or xml is not None:
            return self(xml, filename)
        else:
            return self

    def __call__(self, xml=None, filename=None):
        if filename is not None:
            self.tree = parse(filename, parser=self._parser)
            self.root = self.tree.getroot()
        elif xml is not None:
            self.tree = fromstring(xml, parser=self._parser)
            self.root = self.tree
        else:
            raise Exception("DT called without arguments")
        return self.__end__(self._recurse_node(self.root))

    def __pcdata__(self, text):
        """Method called to process text nodes. If you override it, call its superclass method to
           guarantee empty and strip options to be honored"""
        if not self._empty and str.isspace(text):
            return None
        if self._strip:
            text = text.strip()
        return text

    def __default__(self, element):
        """Default handler for XML elements, when no specific handler is defined"""
        return element.xml

    def __end__(self, result):
        """Handler called after DT process, so it can be used for final processing tasks"""
        return result

    def __comment__(self, text):
        """Handles comment texts. Returns None by default"""
        return None

    def _recurse_node(self, element):
        # copy attributes, so we can store whatever object we want
        if element.tag is Comment:
            comment = self.__comment__(element.text)
            return comment
        else:
            if element.tag in self._alias:
                tag_handler = self._alias[element.tag]
            else:
                tag_handler = getattr(self, element.tag, self.__default__)
                if not callable(tag_handler):
                    tag_handler = self.__default__

            tag_type = self._types.get(element.tag, "string")

            if tag_type == "zero":  # ignore and skip children
                return ""

            # add to path only if required
            self._path.append(Element(element.tag, {**element.attrib}, None, self))

            # given an element, process children, returning a list of pairs
            contents = self._dt(element)      # not zero:
            elem = self._path.pop()

            if tag_type == "string":
                c_list = [ele for t, ele in contents if ele]   # FIXME?? Caso do 0 ???
                elem.c = self.__join__(c_list)

            elif tag_type == "list":
                elem.c = [e for t, e in contents]

            elif tag_type == "map":
                elem.c = {t: e for t, e in contents}

            elif tag_type == "mmap":
                raux = {}
                for t, e in contents:
                    if t in raux:
                        raux[t].append(e)
                    else:
                        raux[t] = [e]
                elem.c = raux

            else:
                raise Exception(f"Element type not recognized: {tag_type}")

            return tag_handler(elem)

    def _dt(self, tree):
        children = []    # agora retorna lista de pares (tag, conte)*
        if tree.text:
            r = self.__pcdata__(tree.text)
            if r:  # should we check for None or "" ?
                children.append(("-pcdata", r))

        for child in tree:
            children.append((child.tag, self._recurse_node(child)))
            if child.tail:
                r = self.__pcdata__(child.tail)
                if r:  # should we check for None or "" ?
                    children.append(("-pcdata", r))
        return children

    def __join__(self, child):
        return str.join("", [str(x) for x in child])

    def __getitem__(self, item):
        dt = {
            "path": self._path,
            "father": self._father,
            "gfather": self._gfather,
            "root": self._root,
            "in_context": lambda: lambda x: self._in_context(x)  # lovely!
        }
        if item in dt:
            return dt[item]()
        else:
            return None

    def _in_context(self, tag):
        return len([e for e in self._path if e.tag == tag]) > 0

    def path(self):
        return self._path

    def _father(self):
        return None if len(self._path) < 1 else self._path[-1]

    def _gfather(self):
        return None if len(self._path) < 2 else self._path[-2]

    def _root(self):
        return None if len(self._path) < 1 else self._path[0]



import pytest
import sys
from xmldt import XmlDt


def test_simple_file():
    class T1 (XmlDt):
        pass

    t1 = T1(strip=True, empty=False)
    assert t1(filename="tests/test1.xml") == """<root>text<body>more text</body>text</root>"""


def test_raise():
    class T1 (XmlDt):
        pass

    t1 = T1()
    with pytest.raises(Exception, match="DT called without arguments"):
        t1()


def test_simple_comment():
    class T1 (XmlDt):
        def __comment__(self, text):
            return text

    t1 = T1(strip=True, empty=False)
    assert t1(filename="tests/test1.xml") == """<root>text A comment <body>more text</body>text</root>"""


def test_simple_types():
    class T1 (XmlDt):
        _types = {"list": "list"}

        def list(self, e):
            assert type(e.c) is list
            assert len(e.c) == 3

        def item(self, e):
            return e.tag

    t1 = T1(strip=True, empty=False)
    t1(filename="tests/test2.xml")


def test_simple_types2():
    class T1 (XmlDt):
        _types = {"list": "list", "tests": "mmap", "submap" : "map"}

        def __default__(self, element):
            return element.c

        def list(self, e):
            assert type(e.c) is list
            assert len(e.c) == 3
            return e.c

        def submap(self, e):
            assert type(e.c) is dict
            assert "zbr" in e.c
            assert e.c["zbr"] == "Zbr"

        def tests(self, e):
            assert type(e.c) is dict
            assert len(e.c["aut"]) == 2
            assert len(e.c["list"][0]) == 3

        def item(self, e):
            return e.tag

    t1 = T1(strip=True, empty=False)
    t1(filename="tests/test3.xml")


def test_zero_type():
    class T1 (XmlDt):
        _types = {"list": "zero"}

        def __default__(self, e):
            return e.xml

    t1 = T1(strip=True, empty=False)
    assert t1(filename="tests/test2.xml") == "<tests/>"


def test_unknown_type():
    class T1 (XmlDt):
        _types = {"list": "junk"}

    t1 = T1(strip=True, empty=False)
    with pytest.raises(Exception):
        t1(filename="tests/test3.xml")



from xmldt import XmlDt
from xmldt.htmldt import HtmlDt


# 1. Test standard XmlDt with html flag
def test_xmldt_html_flag():
    class MyHtmlParser (XmlDt):
        def html(self, element):
            assert True, "Called"

    parser = MyHtmlParser(html=True)
    parser("tests/python.html")


# 2. Test HtmlDt with object
def test_htmldt_with_object():
    class MyHtmlParser2 (HtmlDt):
        def html(self, e):
            assert True, "Called"

    parser = MyHtmlParser2()
    parser("tests/python.html")


# 3. Test HtmlDt with direct call
def test_htmldt_with_direct_call():
    class MyHtmlParser3 (HtmlDt):
        def html(self, e):
            assert True, "Called"

    MyHtmlParser3(filename="tests/python.html")




from xmldt.skel import Skel, build_skel
# import sys
import re


def test_build_skel():
    skel = build_skel("tests/test1.xml")

    # print(skel, file=sys.stderr)
    assert type(skel) is str
    assert re.search(r"def __default__", skel)
    assert re.search(r"# def body", skel)
    assert re.search(r"# def root", skel)


